//********************************************
//
// ChroQ Controller for MSP430F2013
//
//********************************************


#include <msp430x20x3.h>

/*************** For UART *****************************************************************/
#define UART_BIT_TIME           208 /** Period of getting UART */
#define UART_BIT_TIME_0_5       104 /** Period of switching from capture mode to compare mode */
#define UART_WAIT_TO_SEND     10000 /** Wait before it starts to send UART */


/*************** For IR ******************************************************************/
#define IR_CCR1         13      // Period of sending IR for CCR1
#define IR_CCR0         26      // Period of sending IR for CCR0


#define PULSE_HI_0             0.65      // Pulse length of High 0 (msec)
#define PULSE_HI_1             1.05      // Pulse length of High 1 (msec)
#define PULSE_LOW_0            0.35      // Pulse length of Low 0 (msec)
#define PULSE_LOW_1            0.35      // Pulse length of Low 1 (msec)

#define ALL_DURATION             82      // Pulse length of one instruction (msec)
#define START_DURATION         2080      // Pulse length of start bit (usec)
#define REPEAT_COUNT              1      // Number of repeat to send at one instruction

#define KEY_FORWARD              'w'     // FORWARD KEY
#define KEY_BACK                 'z'     // BACK KEY
#define KEY_LEFT                 'a'     // LEFT KEY
#define KEY_RIGHT                's'     // RIGHT KEY
#define KEY_DASH_F               'r'     // DASH FORWARD KEY
#define KEY_DASH_B               'c'     // DASH BACK KEY

#define WDT_PERIOD       (0.064 * 4)     // Period of WDT (msec)

/****************** Option *******************************/
#define MODE_NO_BACK     1 /** No back mode */
#define MODE_D_LR        1 /** Dash foward left or right mode */


/** Channel of Q-Steer */
enum CH{
  CH_A,               // Channel A
  CH_B,               // Channel B
  CH_C,               // Channel C
  CH_D                // Channel D
};

/**  Signal state */
enum STATE{
  STOP,                   // STOP
  FORWARD,                // FORWARD
  BACK,                   // BACK
  LEFT,                   // LEFT
  RIGHT,                  // RIGHT
  DASH_FORWARD,           // DASH FORWARD
  FORWARD_LEFT,           // FORWARD LEFT
  FORWARD_RIGHT,          // FORWARD RIGHT
  DASH_F_L,               // DASH FORWARD LEFT
  DASH_F_R,               // DASH FORWARD RIGHT
  BACK_LEFT,              // BACK LEFT
  BACK_RIGHT,             // BACK RIGHT
  DASH_BACK,              // DASH BACK
  DASH_B_L,               // DASH BACK LEFT
  DASH_B_R,               // DASH BACK RIGHT
  END                     // END
};

// FRAME = 
// {STATE}{INSTRUCTION,ALL_DURATION,START_DURATION,REPEAT_COUNT}
int FRAME[15];

/** Received signals */
unsigned char RX_data;
unsigned char BUF_RX_data;
unsigned int RX_bit_count;

/** State of last instruction */
unsigned int State;
/** WDT interrupt counter */
unsigned int wdt_count;

unsigned int end_flag=0;
unsigned int capture_flag=0;

/******************** Prototype ***********************************/
static void initUART(void);
static void initFrame(char ch);
static void start_uart_rx(void);

static void start_ir(void);
static void setCarrier(void);

static void send(char state);
static void sendIR(int frame);

inline void makeState();

void main(void)
{
  // initialize WDT (0.064 msec interval timer)
  WDTCTL = WDT_MDLY_0_064;
  
  DCOCTL = CALDCO_1MHZ;
  BCSCTL1 = CALBC1_1MHZ;

  // initialize UART
  initUART();
    
  // initialize IR frames
  initFrame(CH_C);
  
  
  while(1){
    //receive UART
    start_uart_rx();
    //start sending IR
    start_ir();
  }
  
}

/**
 * Initialize UART
 */
static void initUART(void)
{
  // P1.1 periferal mode
  P1SEL |= 0x02;
  // P1.1 input
  P1DIR &= ~0x02;

}


/**
 *  Start receiving signals at UART
 */
static void start_uart_rx(void)
{
  // reset bit count for RX
  RX_bit_count = 8;
  // enable interrupt for RX
  CCTL0 = SCS + CM1 + CCIE + CAP;
  TACTL = TASSEL_2 + MC_2;    
  _BIS_SR(GIE);
}

/**
 * TimerA_2 interrupt service for UART
 */
#pragma vector=TIMERA0_VECTOR
__interrupt void uart(void)
{ 
  CCR0 += UART_BIT_TIME; //208
  
  if(CCTL0 & CAP){//start bit get in capture mode

    // switch from capture mode to compare mode
    CCTL0 &= ~CAP;
    // adjust compare of UART
    CCR0 += UART_BIT_TIME_0_5; //104
    // configure TimerA for UART
    TACTL = TASSEL_2 + MC_2;
    capture_flag=0;
      
  }else{          // when it gets data bit
    // read UART signal
    RX_data = RX_data >> 1;
    
    if(CCTL0 & SCCI) RX_data |= 0x80;

    RX_bit_count--;
      
    // when it finished to read 8 bit signals
    if(RX_bit_count == 0){
      // stop interrupt of UART
      CCTL0 &= ~CCIE;
      BUF_RX_data = RX_data;
      makeState();
      _BIC_SR_IRQ(LPM0_bits);
    }//end of if(RX_bit_count == 0)
  }//end of else  
}//end of uart


inline void makeState(){
  // translate from signals to state
  switch(BUF_RX_data){
    case KEY_FORWARD: // "w"
      State = FORWARD;
    break;
      
    case KEY_BACK:    // "z"
#if MODE_NO_BACK
      State = STOP;
#elif
      State = BACK;
#endif
    break;
      
    case KEY_LEFT:    // "a"
#if MODE_D_LR
      State = DASH_F_L;
#elif
      State = LEFT;
#endif
    break;
      
    case KEY_RIGHT:   // "s"
#if MODE_D_LR
      State = DASH_F_R;
#elif
      State = RIGHT;
#endif
    break;
      
    case KEY_DASH_F:  // "r"
      State = DASH_FORWARD;
    break;
      
    case KEY_DASH_B:  // "c"
#if MODE_NO_BACK
      State = STOP;
#elif
      State = DASH_BACK;
#endif
    break;
      
    default:      // others
      State = STOP;
    break;
  }//end of switch
}//end of makecommand



/**
 * Initialize IR frames
 * 
 * @param ch channel of Q-Steer
 */
static void initFrame(char ch)
{
  int i;
  int instruction;
 
  //P1SEL |= 0x04;
  P1DIR |= 0x04;
  // create frame of IR
  for(i=STOP; i < END; i++){
    instruction = ((ch & 0x03) << 4) | (i & 0x0F);
    FRAME[i] = instruction;
  }
}


/**
 * Start sending IR
 */
static void start_ir(void)
{ 
  // configure carrier frequency
  setCarrier();
  
  // State of last instruction
  send(State);
}

/**
 * Configure carrier frequency
 */
static void setCarrier(void)
{
  // configure compare of IR
  CCTL1 = OUTMOD_7;
    //=====================
    // CCTL0 0164h (default 0)
    // TimerA_2 Capture/Compare control 1
    //=====================
    // 0000 0000 1110 0000b
    // 00~~ ~~~~ ~~~~ ~~~~  (CM)    not capture
    // ~~00 ~~~~ ~~~~ ~~~~  (CCIS)
    // ~~~~ 0~~~ ~~~~ ~~~~  (SCS)   not synchronize
    // ~~~~ ~0~~ ~~~~ ~~~~  (SCCI)
    // ~~~~ ~~0~ ~~~~ ~~~~  (Reserved)
    // ~~~~ ~~~0 ~~~~ ~~~~  (CAP)
    // ~~~~ ~~~~ 111~ ~~~~  (OUTMOD)  reset/set mode (CCR0 set, CCR1 reset)
    // ~~~~ ~~~~ ~~~0 ~~~~  (CCIE)    disable CCIFG interrupt
    // ~~~~ ~~~~ ~~~~ 0~~~  (CCI)
    // ~~~~ ~~~~ ~~~~ ~0~~  (OUT)   reset OUT-F/F of CCR1
    // ~~~~ ~~~~ ~~~~ ~~0~  (COV)   detection capture over flow
    // ~~~~ ~~~~ ~~~~ ~~~0  (CCIFG)
  
  // configure compares
  CCR0 = IR_CCR0;
  CCR1 = IR_CCR1;
  
  // configure TimerA for IR
  TACTL = TASSEL_2 + MC_1;
    //=====================
    // TACTL 0160h (default 0)
    // TimerA_2 Control
    //=====================
    // 0000 0010 0001 0000b  
    // 0000 00~~ ~~~~ ~~~~  (Reserved)
    // ~~~~ ~~10 ~~~~ ~~~~  (TASSEL)  clock source : SMCLK
    // ~~~~ ~~~~ 00~~ ~~~~  (ID)    frequency dividing percentage of clock : 1
    // ~~~~ ~~~~ ~~01 ~~~~  (MC)    up mode
    // ~~~~ ~~~~ ~~~~ 0~~~  (Reserved)
    // ~~~~ ~~~~ ~~~~ ~0~~  (TACLR)   reset counter
    // ~~~~ ~~~~ ~~~~ ~~0~  (TAIE)    disable TAIFG interrupt 
    // ~~~~ ~~~~ ~~~~ ~~~0  (TAIFG)   interrupt flag
    
}

/**
 * Send instruction
 * 
 * @param state Received signals
 */
static void send(char state)
{
  int frame;
  
  // translate from state to frame
  switch(state){
  case FORWARD:
    frame = FRAME[FORWARD];
    break;
  case BACK:
    frame = FRAME[BACK];
    break;
  case RIGHT:
    frame = FRAME[RIGHT];
    break;
  case LEFT:
    frame = FRAME[LEFT];
    break;
  case DASH_FORWARD:
    frame = FRAME[DASH_FORWARD];
    break;
  case FORWARD_LEFT:
    frame = FRAME[FORWARD_LEFT];
    break;
  case FORWARD_RIGHT:
    frame = FRAME[FORWARD_RIGHT];
    break;
  case DASH_F_L:
    frame = FRAME[DASH_F_L];
    break;
  case DASH_F_R:
    frame = FRAME[DASH_F_R];
    break;
  case BACK_LEFT:
    frame = FRAME[BACK_LEFT];
    break;
  case BACK_RIGHT:
    frame = FRAME[BACK_RIGHT];
    break;
  case DASH_BACK:
    frame = FRAME[DASH_BACK];
    break;
  case DASH_B_L:
    frame = FRAME[DASH_B_L];
    break;
  case DASH_B_R:
    frame = FRAME[DASH_B_R];
    break;
  case STOP:
    frame = FRAME[STOP];
    break;
  }
  

  do{
    sendIR(frame);
  }while(capture_flag);
  capture_flag=1;
}

/**
 * Send IR
 * 
 * @param frame send frame
 */
static void sendIR(int frame)
{
  int i, j, k;
  int term_period;
  /*************************************/
  /* configure frame                   */
  /*************************************/
 
  // configure IR parameters
  char   instruction     = (char)(frame & 0x3F);
  int    all_duration    = ALL_DURATION;
  double start_duration  = START_DURATION / 1000;
  int    repeat_count    = REPEAT_COUNT;
  
  // calc number of roops
  int roop_rest=0;
  int roop_start =   start_duration / WDT_PERIOD;
  int roop_all =       all_duration / WDT_PERIOD;
  int roop_true =        PULSE_HI_1 / WDT_PERIOD;
  int roop_false =       PULSE_HI_0 / WDT_PERIOD;
  roop_false += 1;

  // enable WDT interrupt
  IE1 |= WDTIE;


  /*************************************/
  /* send IR                           */
  /*************************************/
  // repeat for repeat_count
  for(i=0; i < repeat_count; i++){
    
    // roop number of rest
    roop_rest = roop_all;
    
    // reset WDT interrupt counter
    wdt_count = 0;
    
    // start bit send
    for(k=0; k < roop_start; k++){
      P1SEL |= 0x04;
      _BIS_SR(LPM0_bits + GIE);
    }
    
    // stop carrier
    P1SEL &= ~0x04;
    
    roop_rest -= roop_start;
    
    // 6bit send 
    for(j=0; j < 6; j++){
      
      // wait for sending low bit
      _BIS_SR(LPM0_bits + GIE);
      
      if(instruction & 0x20) term_period = roop_true;
      else term_period = roop_false;
        
      // start sending carrier for datas
      for(k=0; k < term_period; k++){
        _BIS_SR(LPM0_bits + GIE);
        P1SEL |= 0x04;
      }

      // stop sending carrier
      P1SEL &= ~0x04;

      // shift instruction
      instruction <<= 1;

      roop_rest -= term_period;
    }
    
    // wait for going on all duration
    for(k=0; k < roop_rest; k++){
      _BIS_SR(LPM0_bits + GIE);
    }
  }
  
  // disable WDT interrupt
  IE1 &= ~WDTIE;
}

/**
 * WDT interrupt service for 0.064 msec interval timer
 */
#pragma vector=WDT_VECTOR
__interrupt void watchdog_timer(void)
{
  wdt_count++;
  
  if(wdt_count >= 4){   // if it goes on 0.26 msec (0.064 * 4)
    wdt_count = 0;
    // restart IR rootine
    _BIC_SR_IRQ(LPM0_bits);
  }
}



