//********************************************
//
// Chro-Q Controller for MSP430F2013
//
//********************************************


#include <msp430x20x3.h>

/*************** UART定数 ****************************************************************/
#define UART_BIT_TIME           208 // UART受信間隔
#define UART_BIT_TIME_0_5       104 // UART受信間隔/2


/*************** 赤外線送信定数 ************************************************************/
#define IR_CCR1         13      // 赤外線送信間隔(CCR1)
#define IR_CCR0         26      // 赤外線送信間隔(CCR0)


#define PULSE_HI_0             0.65      // High0のパルス長 (msec)
#define PULSE_HI_1             1.05      // High1のパルス長 (msec)
#define PULSE_LOW_0            0.35      // Low0のパルス長 (msec)
#define PULSE_LOW_1            0.35      // Low1のパルス長 (msec)

#define ALL_DURATION             82      // 全体長 (msec)
#define START_DURATION         2080      // スタート長 (usec)
#define REPEAT_COUNT              1      // リピート数

#define KEY_FORWARD              'w'     // 前進キー
#define KEY_BACK                 'z'     // 後退キー
#define KEY_LEFT                 'a'     // 左折キー
#define KEY_RIGHT                's'     // 右折キー
#define KEY_DASH_F               'r'     // 加速前進キー
#define KEY_DASH_B               'c'     // 加速後退キー

#define WDT_PERIOD       (0.064 * 4)     // WDT割り込み間隔*4 (msec)

/****************** Option *******************************/
#define MODE_NO_BACK     1 // 1で後退をオフにする
#define MODE_D_LR        1 // 1で[左折/右折]が[左折前進/右折前進]になる


/** Q-Steerのチャンネル */
enum CH{
  CH_A,               // Channel A
  CH_B,               // Channel B
  CH_C,               // Channel C
  CH_D                // Channel D
};

/** Q-Steerの命令 */
enum STATE{
  STOP,                   // 停止
  FORWARD,                // 前進
  BACK,                   // 後退
  LEFT,                   // 左折
  RIGHT,                  // 右折
  DASH_FORWARD,           // 加速全身
  FORWARD_LEFT,           // 左折前進
  FORWARD_RIGHT,          // 右折前進
  DASH_F_L,               // 左折加速全身
  DASH_F_R,               // 右折加速全身
  BACK_LEFT,              // 左折後退
  BACK_RIGHT,             // 右折後退
  DASH_BACK,              // 加速後退
  DASH_B_L,               // 左折加速後退
  DASH_B_R,               // 右折加速後退
  END                     // END
};

// 赤外線送信フレーム
int FRAME[15];

/** 受信データ */
unsigned char RX_data;
unsigned char BUF_RX_data;
unsigned int RX_bit_count;

/** 命令の種類 */
unsigned int State;
/** WDTの割り込み回数 */
unsigned int wdt_count;

unsigned int end_flag=0;
unsigned int capture_flag=0;

/******************** プロトタイプ宣言 ***********************************/
static void initUART(void);
static void initFrame(char ch);
static void start_uart_rx(void);

static void start_ir(void);
static void setCarrier(void);

static void send(char state);
static void sendIR(int frame);

inline void makeState();

void main(void)
{
  // WDTの初期化 (0.064msec インターバルタイマ)
  WDTCTL = WDT_MDLY_0_064;
  
  DCOCTL = CALDCO_1MHZ;
  BCSCTL1 = CALBC1_1MHZ;

  // UARTの初期化
  initUART();
    
  // 赤外線FRAMEの初期化
  initFrame(CH_C);
  
  
  while(1){
    //命令の受信
    start_uart_rx();
    //赤外線の送信
    start_ir();
  }
  
}

/**
 * UARTの初期化
 */
static void initUART(void)
{
  // P1.1をペリフェラルに設定
  P1SEL |= 0x02;
    //=====================
    // P1SEL 026h
    //=====================
    // 0000 0010b
    // 0~~~ ~~~~  (P1.7)
    // ~0~~ ~~~~  (P1.6)
    // ~~0~ ~~~~  (P1.5)
    // ~~~0 ~~~~  (P1.4)
    // ~~~~ 0~~~  (P1.3)
    // ~~~~ ~0~~  (P1.2)
    // ~~~~ ~~1~  (P1.1)  Periferal
    // ~~~~ ~~~0  (P1.0)
  
  // P1.1を入力に設定
  P1DIR &= ~0x02;
    //=====================
    // P1SEL 026h
    //=====================
    // 0010 0010b
    // 0~~~ ~~~~  (P1.7)
    // ~0~~ ~~~~  (P1.6)
    // ~~0~ ~~~~  (P1.5)  Periferal
    // ~~~0 ~~~~  (P1.4)
    // ~~~~ 0~~~  (P1.3)
    // ~~~~ ~0~~  (P1.2)
    // ~~~~ ~~0~  (P1.1)
    // ~~~~ ~~~0  (P1.0)
}


/**
 * 命令の受信を開始する
 */
static void start_uart_rx(void)
{

  // 受信ビット数を初期化
  RX_bit_count = 8;
  
  // キャプチャ/コンペア制御レジスタ0
  CCTL0 = SCS + CM1 + CCIE + CAP;
    //=====================
    // CCTL0 0162h (default 0)
    // TimerA_2 Capture/Compare control 0
    //=====================
    // 1000 0001 0001 0000b
    // 10~~ ~~~~ ~~~~ ~~~~  (CM)     キャプチャタイミング: 立下りエッジ
    // ~~00 ~~~~ ~~~~ ~~~~  (CCIS)   CCI0A(P1.1)
    // ~~~~ 0~~~ ~~~~ ~~~~  (SCS)    synchronize timer clock
    // ~~~~ ~0~~ ~~~~ ~~~~  (SCCI)
    // ~~~~ ~~0~ ~~~~ ~~~~  (Reserved)
    // ~~~~ ~~~1 ~~~~ ~~~~  (CAP)    キャプチャモード
    // ~~~~ ~~~~ 000~ ~~~~  (OUTMOD)
    // ~~~~ ~~~~ ~~~1 ~~~~  (CCIE)   割り込みを許可
    // ~~~~ ~~~~ ~~~~ 0~~~  (CCI)
    // ~~~~ ~~~~ ~~~~ ~0~~  (OUT)
    // ~~~~ ~~~~ ~~~~ ~~0~  (COV)
    // ~~~~ ~~~~ ~~~~ ~~~0  (CCIFG)
  
  CCR0=0;
  CCR1=0;
  
  // TimerA_2の設定
  TACTL = TASSEL_2 + MC_2;
    //=====================
    // TACTL 0160h (default 0)
    // TimerA_2 Control
    //=====================
    // 0000 0010 0010 0000b  
    // 0000 00~~ ~~~~ ~~~~  (Reserved)
    // ~~~~ ~~10 ~~~~ ~~~~  (TASSEL)  クロックソース: SMCLK
    // ~~~~ ~~~~ 00~~ ~~~~  (ID)      
    // ~~~~ ~~~~ ~~10 ~~~~  (MC)      連続モード
    // ~~~~ ~~~~ ~~~~ 0~~~  (Reserved)
    // ~~~~ ~~~~ ~~~~ ~0~~  (TACLR)   reset counter
    // ~~~~ ~~~~ ~~~~ ~~0~  (TAIE)    disable TAIFG interrupt 
    // ~~~~ ~~~~ ~~~~ ~~~0  (TAIFG)   interrupt flag
    
  _BIS_SR(GIE);
}

/**
 * TimerA_2割り込みサービスルーチン
 */
#pragma vector=TIMERA0_VECTOR
__interrupt void uart(void)
{ 
  CCR0 += UART_BIT_TIME; //208
  
  if(CCTL0 & CAP){// スタートビットを検出(キャプチャモード)
      
    // コンペアモードに設定
    CCTL0 &= ~CAP;
    // 波形の真ん中を取得するようにする
    CCR0 += UART_BIT_TIME_0_5; //104
    // TimerA_2をUART用に設定
    TACTL = TASSEL_2 + MC_2;
    capture_flag=0;
      
  }else{          // データを受信
    
    RX_data = RX_data >> 1;
    
    // 1bit読み込む
    if(CCTL0 & SCCI){
      RX_data |= 0x80;
    }
      
    RX_bit_count--;
      
    // 8bit読み込んだとき
    if(RX_bit_count == 0){
      // Timer_A2割り込みを停止する
      CCTL0 &= ~CCIE;
      BUF_RX_data = RX_data;
      makeState();
      
      _BIC_SR_IRQ(LPM0_bits);
      
    }//end of if(RX_bit_count == 0)
  }//end of else  
}//end of uart


inline void makeState(){
  
  // 受信したデータを命令に変換
  switch(BUF_RX_data){
    case KEY_FORWARD: // "w"
      State = FORWARD;
    break;
      
    case KEY_BACK:    // "z"
#if MODE_NO_BACK
      State = STOP;
#elif
      State = BACK;
#endif
    break;
      
    case KEY_LEFT:    // "a"
#if MODE_D_LR
      State = DASH_F_L;
#elif
      State = LEFT;
#endif
    break;
      
    case KEY_RIGHT:   // "s"
#if MODE_D_LR
      State = DASH_F_R;
#elif
      State = RIGHT;
#endif
    break;
      
    case KEY_DASH_F:  // "r"
      State = DASH_FORWARD;
    break;
      
    case KEY_DASH_B:  // "c"
#if MODE_NO_BACK
      State = STOP;
#elif
      State = DASH_BACK;
#endif
    break;
      
    default:      // others
      State = STOP;
    break;

  }//end of switch
}//end of makecommand



/**
 * 赤外線送信フレームの初期化
 * 
 * @param ch Q-Steerのチャンネル
 */
static void initFrame(char ch)
{
  int i;
  int instruction;
  
  P1DIR |= 0x04;
  // 赤外線送信フレームを生成
  for(i=STOP; i < END; i++){
    instruction = ((ch & 0x03) << 4) | (i & 0x0F);
    FRAME[i] = instruction;
  }
}


/**
 * 赤外線の送信を開始する
 */
static void start_ir(void)
{ 
  // キャリア周波数の設定
  setCarrier();
  
  // 命令を送信する
  send(State);
}

/**
 * キャリア周波数の設定
 */
static void setCarrier(void)
{
  // キャプチャ/コンペア制御レジスタ1
  CCTL1 = OUTMOD_7;
    //=====================
    // CCTL0 0164h (default 0)
    // TimerA_2 Capture/Compare control 1
    //=====================
    // 0000 0000 1110 0000b
    // 00~~ ~~~~ ~~~~ ~~~~  (CM)      not capture
    // ~~00 ~~~~ ~~~~ ~~~~  (CCIS)
    // ~~~~ 0~~~ ~~~~ ~~~~  (SCS)     not synchronize
    // ~~~~ ~0~~ ~~~~ ~~~~  (SCCI)
    // ~~~~ ~~0~ ~~~~ ~~~~  (Reserved)
    // ~~~~ ~~~0 ~~~~ ~~~~  (CAP)
    // ~~~~ ~~~~ 111~ ~~~~  (OUTMOD)  reset/set mode (CCR0 set, CCR1 reset)
    // ~~~~ ~~~~ ~~~0 ~~~~  (CCIE)    disable CCIFG interrupt
    // ~~~~ ~~~~ ~~~~ 0~~~  (CCI)
    // ~~~~ ~~~~ ~~~~ ~0~~  (OUT)     reset OUT-F/F of CCR1
    // ~~~~ ~~~~ ~~~~ ~~0~  (COV)     detection capture over flow
    // ~~~~ ~~~~ ~~~~ ~~~0  (CCIFG)
  
  // キャプチャ/コンペアレジスタの初期化
  CCR0 = IR_CCR0;
  CCR1 = IR_CCR1;
  
  // TimerA_2の設定
  TACTL = TASSEL_2 + MC_1;
    //=====================
    // TACTL 0160h (default 0)
    // TimerA_2 Control
    //=====================
    // 0000 0010 0001 0000b  
    // 0000 00~~ ~~~~ ~~~~  (Reserved)
    // ~~~~ ~~10 ~~~~ ~~~~  (TASSEL)  クロックソース: SMCLK
    // ~~~~ ~~~~ 00~~ ~~~~  (ID)
    // ~~~~ ~~~~ ~~01 ~~~~  (MC)      アップモード
    // ~~~~ ~~~~ ~~~~ 0~~~  (Reserved)
    // ~~~~ ~~~~ ~~~~ ~0~~  (TACLR)   reset counter
    // ~~~~ ~~~~ ~~~~ ~~0~  (TAIE)    disable TAIFG interrupt 
    // ~~~~ ~~~~ ~~~~ ~~~0  (TAIFG)
    
}

/**
 * 命令を送信する
 * 
 * @param state 受信した命令
 */
static void send(char state)
{
  int frame;
  
  // 命令から送信フレームを決定する
  switch(state){
  case FORWARD:
    frame = FRAME[FORWARD];
    break;
  case BACK:
    frame = FRAME[BACK];
    break;
  case RIGHT:
    frame = FRAME[RIGHT];
    break;
  case LEFT:
    frame = FRAME[LEFT];
    break;
  case DASH_FORWARD:
    frame = FRAME[DASH_FORWARD];
    break;
  case FORWARD_LEFT:
    frame = FRAME[FORWARD_LEFT];
    break;
  case FORWARD_RIGHT:
    frame = FRAME[FORWARD_RIGHT];
    break;
  case DASH_F_L:
    frame = FRAME[DASH_F_L];
    break;
  case DASH_F_R:
    frame = FRAME[DASH_F_R];
    break;
  case BACK_LEFT:
    frame = FRAME[BACK_LEFT];
    break;
  case BACK_RIGHT:
    frame = FRAME[BACK_RIGHT];
    break;
  case DASH_BACK:
    frame = FRAME[DASH_BACK];
    break;
  case DASH_B_L:
    frame = FRAME[DASH_B_L];
    break;
  case DASH_B_R:
    frame = FRAME[DASH_B_R];
    break;
  case STOP:
    frame = FRAME[STOP];
    break;
  }
  

  do{
    sendIR(frame);
  }while(capture_flag);
  capture_flag=1;
}

/**
 * 赤外線を送信する
 * 
 * @param frame 送信フレーム
 */
static void sendIR(int frame)
{
  int i, j, k;
  int term_period;
  
  /*************************************/
  /* フレームの設定                                               */
  /*************************************/
 
  // 赤外線の送信パラメータを設定
  char   instruction     = (char)(frame & 0x3F);
  int    all_duration    = ALL_DURATION;
  double start_duration  = START_DURATION / 1000;
  int    repeat_count    = REPEAT_COUNT;
  
  // ループ回数を算出
  int roop_rest=0;
  int roop_start =   start_duration / WDT_PERIOD;
  int roop_all =       all_duration / WDT_PERIOD;
  int roop_true =        PULSE_HI_1 / WDT_PERIOD;
  int roop_false =       PULSE_HI_0 / WDT_PERIOD;
  roop_false += 1;

  // WDT割り込みを許可
  IE1 |= WDTIE;


  /*************************************/
  /* send IR                           */
  /*************************************/
  // リピート回数分赤外線を送信する。
  for(i=0; i < repeat_count; i++){
    
    roop_rest = roop_all;
    
    // WDT割り込みの回数を初期化
    wdt_count = 0;
    
    // スタートビットを送信
    for(k=0; k < roop_start; k++){
      P1SEL |= 0x04;
      _BIS_SR(LPM0_bits + GIE);
    }
    
    // キャリアを停止
    P1SEL &= ~0x04;
    
    roop_rest -= roop_start;
    
    // 命令を送信
    for(j=0; j < 6; j++){
      
      // Low送信中はキャリアを停止する
      _BIS_SR(LPM0_bits + GIE);
      
      if(instruction & 0x20) term_period = roop_true;
      else term_period = roop_false;
        
      // キャリアの送信を開始する
      for(k=0; k < term_period; k++){
        _BIS_SR(LPM0_bits + GIE);
        P1SEL |= 0x04;
      }

      // キャリアを停止
      P1SEL &= ~0x04;

      // 命令をシフトする
      instruction <<= 1;

      roop_rest -= term_period;
    }
    
    // 全体長分待機する
    for(k=0; k < roop_rest; k++){
      _BIS_SR(LPM0_bits + GIE);
    }
  }
  
  // WDT割り込みを停止
  IE1 &= ~WDTIE;
}

/**
 * WDT割り込みサービスルーチン(0.064msec インターバルタイマ)
 */
#pragma vector=WDT_VECTOR
__interrupt void watchdog_timer(void)
{
  wdt_count++;
  
  if(wdt_count >= 4){   // 0.26msec経過したとき
    wdt_count = 0;
    // 赤外線送信処理を再開する
    _BIC_SR_IRQ(LPM0_bits);
  }
}



